/*
 * Population Class Definition Header File
 *
 * Pop    is an instance of a population group
 * PPop   is a pointer to an instance of a population group
 *
 * John Dunkelberg
 * created: 22 Feb 1994
 */

#include "problem.H"
#include "chromosone.H"

extern int DEBUG;

class Pop
{
  int generation;
  long pop_size;
  
  // The chromosones in the population
  Chromosone *first_chrom;
  Chromosone *last_chrom;
  Chromosone *best_chrom;
  Chromosone *tplate;

  // The mesh and machine definitions
  Problem problem;

public:
// Construct and Destruct Functions
  Pop();         // constructor (doesn't assign anything serious)
  ~Pop();        // destructor (kill all chromosones)

  int Init_File(fstream);   // Initializes the problem desc.
  int Init_Tplate(void);    // Assigns a basic tplate
  long Initialize(int);     // Assigns values to the population

// Private Info Functions
  long Pop_Size(void)
    { return pop_size; }  // returns the size of the population
  double Get_Best_Fit(void);  // returns fitness of best chrom
  void Show_Best(int);        // Show information on the best chromosone
  void Show_Best();
  
// Selection Functions
  void Fit_All();
  int Shuffle();          // randomly shuffle the population
  long Init_Select(long);
  long Select(long);
  long Juxta(long);

// Internal Functions
  // Permutate the forms & create the chromosones
  void Permutate(Chromosone *,long *,int *,int,Gene *,int,int);
  double Statistics();  // returns average fitness value of pop
  void Show(int);
  void Show(void);
};

typedef Pop *PPop;

/*
*  functions for the population class
*
*/

// Construct a void population
Pop::Pop()
{
  first_chrom = last_chrom = best_chrom = tplate = (Chromosone *) NULL;
  pop_size = 0;
};

// Destroy all chromosones and values
Pop::~Pop()
{
  if (first_chrom != (Chromosone *) NULL)
    {
      first_chrom->Kill();
      delete first_chrom;
    }

  if (tplate != (Chromosone *) NULL)
    delete tplate;
  
}

// Initialize the problem definition
int Pop::Init_File(fstream in_fp)
{
  // initialize the problem
  problem.Initialize(in_fp);
  problem.Show(DEBUG);
  
  in_fp.close();
  return(1);
}

// Initialize the tplate
int Pop::Init_Tplate()
{
  if (problem.mach_size == 0)
    {
      cout <<"Pop::Init_Tplate Problem must be defined prior to tplate\n";
      cout <<"Initialization...\n";
      exit(0);
    }

  tplate = new Chromosone;
  int cur_gene, cur_allele;
  Gene *fgene, *lgene, *eugene;
  {
    fgene = new Gene;
    fgene->Assign(0,0);
    lgene = fgene;
  }
  for(cur_gene=1;(cur_gene < problem.size);cur_gene++)
    {
      eugene = new Gene;
      eugene->Assign(cur_gene,cur_allele);
      lgene->Connect(eugene);
      lgene = eugene;

      cur_allele = (++cur_allele)%(problem.mach_size);
    }

  cout << "And now assign...\n";

  tplate->Assign((long) 0,fgene);

  cout << "Assigned" << endl;

  if (DEBUG > 50)
    tplate->Show(DEBUG);

  return(1);
}

// Initialize sets up the original population (all BBs)
// it returns the new population_size
long Pop::Initialize(int order)
{
  long real_pop_size = 0;      // actual pop_size count
  int prob_size, mach_size;    // problem size variables
  long pool_size;              // size of the gene pool
  Gene *pool;                  // the gene pool
  Chromosone *init_pop;        // initial population
  int count;                   // dummy loop variable

  prob_size = problem.Get_Size();
  mach_size = problem.Get_Mach_Size();

  // We determine the size gene pool of all the possible genes
  pool_size = ((long) prob_size) * ((long) mach_size);

  if (DEBUG)
    {
      cout << "Gene Pool size: " << pool_size;
      cout << " Mem: " << (pool_size*sizeof(Gene)) << endl;
    }

  if (order > pool_size)
    {
      cout << "It is not possible to create unique chromosones of order\n";
      cout << order << "in this problem, maximum is " << pool_size << endl;
      return(0);
    }

  // Allocate space for the gene pool
  pool = new Gene[pool_size];
  if (pool == (Gene *) NULL)
    {
      cout << "Pop::Initialize was unable to create a gene_pool\n";
      cout << "Attempted Gene Pool size: " << pool_size << " Genes\n";
      cout << "memory: " << (pool_size * sizeof(Gene)) << endl;
      exit(1);
    }

  // Determine the size of the total population
  // pop size is (pool_size choose order)  p_s C ord
  for(count=0,pop_size=1;(count < order);count++)
    pop_size *= (pool_size-count);
  for(count=2;(count <= order);count++)
    pop_size /= count;

  if (DEBUG)
    {
      cout << "Init Population Size: " << pop_size;
      cout << " Mem: " << (pop_size * sizeof(Chromosone)) << endl;
    }

  // Allocate space for the chromosone population
  init_pop = new Chromosone[pop_size];
  if (init_pop == (Chromosone *) NULL)
    {
      cout << "Pop::Initialize was unable to create a initial population\n";
      cout << "Attempted Population size: " << pop_size << " Chroms\n";
      cout << "memory: " << (pop_size * sizeof(Chromosone)) << endl;
      exit(1);
    }

  cout << "init_pop &=" << init_pop;

  // Assign values to the genes in the pool
  unsigned int cur_ps,cur_ms;   // counters
  for(cur_ps=0; (cur_ps < (unsigned int) prob_size) ;cur_ps++)
    for(cur_ms=0; (cur_ms < (unsigned int) mach_size) ;cur_ms++)
      {
	pool[((mach_size * cur_ps)+cur_ms)].Assign(cur_ps,cur_ms);
	if (DEBUG > 50)
	  {
	    cout << "PS: " << cur_ps << " MS: " << cur_ms << endl;
	    pool[((mach_size * cur_ps)+cur_ms)].Show();
	  }
      }

  // create all the chromosones we want
  int *form;                    // gene format for chromosone

  form = new int[order];
  for(count=0;(count<order);count++)   // initialize our form for the
    form[count] = count;               // chromosones

  // recursively make them all
  Permutate(init_pop,&real_pop_size,form,0,pool,pool_size,order);

  if (DEBUG > 10)
    cout << "real_pop_size " << real_pop_size << endl;

  // clean up some memory
  delete pool;
  delete form;
  
  return(real_pop_size);
}

// Permutate 
// This is done recursively, which is NOT the best way when one considers
// that we're going to be storing thousands of functions on the stack
// (But it works... for now)
void Pop::Permutate(Chromosone *pop, long *size,
	       int *form, int level,
	       Gene *pool, int pool_size, int order)
{
  int value, init_val, max_val;
  max_val = pool_size - (order - level);
  if (level == 0)
    init_val = 0;
  else
    init_val = form[level-1]+1;
  for(value=init_val;(value <= max_val);value++)
    {
      form[level] = value;
      if (level < (order-1))
	{
	  Permutate(pop,size,form,(level+1),pool,pool_size,order);
	}
      else
	{
	  // Create the Chromosone
	  int count;
	  Chromosone *new_chrom;

	  // get new space from prev. allocated block, increment *size
	  new_chrom = &(pop[(*size)++]);
	  if (new_chrom == (Chromosone *) NULL)
	    {
	      cout << "Pop::Initialize:Permutate failed to allocate space ";
	      cout << "for new chromosone\n";
	      cout << "Chrom: " << (*size) << endl;
	      exit(1);
	    }

	  if (DEBUG > 50)
	    cout << "new Chrom " << *size << ":";

	  // Creation variables
	  Gene *base_gene;   // the first gene for the Chrom
	  Gene *new_gene;    // newly created gene
	  Gene *last_gene;   // the current addition point for new genes
	  int pool_id;                  // the id of a gene in the pool
	  unsigned int new_gene_id, new_allele;  // the vals of the gene

	  // The base gene (first_gene of chrom)
	  base_gene = new Gene;
	  if (base_gene == (Gene *) NULL)
	    {
	      cout << "Permutate failed to get room for a Gene\n";
	      exit(1);
	    }
	  pool_id = form[0];
	  new_gene_id = pool[pool_id].Gene_id();
	  new_allele = pool[pool_id].Allele();
	  base_gene->Assign(new_gene_id,new_allele);
	  last_gene = base_gene;

	  if (DEBUG > 50)
	    base_gene->Show(DEBUG);

	  // Subsequent Genes (for order > 1)
	  for(count=1;(count<order);count++)
	    {
	      new_gene = new Gene;
	      if (new_gene == (Gene *) NULL)
		{
		  cout << "Permutate failed to get room for a Gene\n";
		  exit(1);
		}

	      pool_id = form[count];
	      new_gene_id = pool[pool_id].Gene_id();
	      new_allele = pool[pool_id].Allele();
	      new_gene->Assign(new_gene_id,new_allele);
	      last_gene->Connect(new_gene);
	      last_gene = last_gene->Next();

	      if (DEBUG > 50)
		new_gene->Show(DEBUG);
	    }

	  // Assign gene_sequence and id to the chromsone
	  new_chrom->Assign((*size),base_gene);

	  if (DEBUG >10)
	    new_chrom->Show(DEBUG);

	  // Add the new chromosone to the population
	  if ((*size) == 1)
	    {
	      // 1st chromosone
	      first_chrom = new_chrom;
	      last_chrom = new_chrom;
	    }
	  else
	    {
	      // nth chromosone
	      last_chrom->Connect(new_chrom);
	      last_chrom = last_chrom->Next();
	    }

	  if (DEBUG > 40)
	    cout << "new Chrom " << (*size) << " added to pop\n";
	}
    }
  return;
}

// TEST function to place fitness on all chromosones
void Pop::Fit_All()
{
  Chromosone *cur_chrom;

  for(cur_chrom = first_chrom;
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next())
    {
      cur_chrom->Fitness(&problem,tplate);
    }
}

// Shuffle the population randomly
int Pop::Shuffle()
{
  // Create Shuffle chain
  Chromosone *sfirst_chrom, *slast_chrom;
  Chromosone *cur_chrom,*tmp_chrom,*tail_chrom,*new_chrom;
  long shuffle_dist;

  long cur_pop;
  sfirst_chrom = first_chrom->Copy();
  slast_chrom = sfirst_chrom;
  for(cur_chrom = first_chrom->Next(), cur_pop = 2;
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next(), cur_pop++)
    {
      // choose a distance to shuffle it in
      shuffle_dist = random() % cur_pop;

      for(tmp_chrom = sfirst_chrom;(shuffle_dist);shuffle_dist--)
	{
	  if (tmp_chrom->Next() != (Chromosone *) NULL)
	    tmp_chrom = tmp_chrom->Next();
	}

      new_chrom = cur_chrom->Copy();

      if (tmp_chrom == sfirst_chrom)
	{
	  // put new element on the head of the shuffle chain
	  new_chrom->Connect(tmp_chrom);
	  sfirst_chrom = new_chrom;
	}
      else
	{
	  // put the chromosone in the middle
	  tail_chrom = tmp_chrom->Disconnect();
	  tmp_chrom->Connect(new_chrom);
	  new_chrom->Connect(tail_chrom);
	}
      
      if (tmp_chrom->Next() == (Chromosone *) NULL)
	{
	  // this is a last in the shuffle chain
	  slast_chrom = tmp_chrom;
	}
      
    }
  // get rid off the old chromosone population
  first_chrom->Kill();

  // put back in the shuffled population
  first_chrom = sfirst_chrom;
  last_chrom = slast_chrom;

  return(1);
}


// Init_Select does the initial selection routine, which does non-shuffle
// tournament selection, without geneic operators
// This assumes that all of the chromosones have been fittness tested
long Pop::Init_Select(long targ_pop_size)
{
  // the new chromosone chain for the population
  Chromosone *nfirst_chrom, *nlast_chrom;
  nfirst_chrom = nlast_chrom = (Chromosone *) NULL;

  long cur_pop;
  Chromosone *first, *second, *fittest;
  first = second = fittest = (Chromosone *) NULL;

  for(cur_pop = 0; (cur_pop < targ_pop_size); cur_pop++)
    {
      // Choose the next two population chromosones
      {
	// Choose first of the two chromosone
	if (first == (Chromosone *) NULL)
	  first = first_chrom;
	else
	  {
	    // Get next unmarked chromosone
	    for(first = first->Next();
		((first != (Chromosone *) NULL) && (first->Get_Mark()));
		first = first->Next());
	    if (first == (Chromosone *) NULL)
	      {
		// We've gone through all the chroms and need more
		first = first_chrom;   // so reset
		first->Unmark_All();   // unmark everything
		Shuffle();             // and shuffle 'em
	      }
	    if (first->Get_Fit() == -1.0)
	      {
		// chromosone has not been fitness tested.. doit
		first->Fitness(&problem, tplate);
	      }
	  }

	if (DEBUG>40)
	  {
	    cout << "first chromosone is: ";
	    first->Show(DEBUG-20);
	  }

	// Choose second chromosone from unmarked chromosones
	// Make sure they meet requirements for thresholding
	for(second = first->Next();
	    ((second != (Chromosone *) NULL) &&
	     (second->Threshold(first)) &&
	     (second->Get_Mark()));
	    second = second->Next());
      } // end of the first,second block (not a loop construct!)

          
      // Determine the best Get_Fitness of them
      if (second == (Chromosone *) NULL)
	{
	  if (DEBUG > 30)
	    cout << "no second found... proceeding\n";

	  // there is no threshold match for first, so pass through?
	  fittest = first->Copy();
	  first->Mark();
	}
      else
	{
	  // mark them as tested
	  first->Mark();
	  second->Mark();

	  if (second->Get_Fit() == -1.0)
	    {
	      // chromosone has not been fitness tested.. doit
	      second->Fitness(&problem, tplate);
	    }
	  
	  if (DEBUG>40)
	    {
	      cout << "second chromosone is: ";
	      second->Show(DEBUG-20);
	    }
	  
	  if (first->Get_Fit() == second->Get_Fit())
	    {
	      if (first->Get_Len() < second->Get_Len())
		fittest = second->Copy();
	      else
		fittest = first->Copy();
	    }
	  else if (first->Get_Fit() > second->Get_Fit())
	    fittest = first->Copy();
	  else
	    fittest = second->Copy();
	}

      // Put the best Get_Fitness in the new population
      if (nfirst_chrom == (Chromosone *) NULL)
	{
	  nfirst_chrom = fittest;
	  nlast_chrom = fittest;
	}
      else
	{
	  nlast_chrom->Connect(fittest);
	  nlast_chrom = nlast_chrom->Next();
	}
    } // end of selection block
  
  // replace current chromosone string with the new chromosone string
  Chromosone *temp_chrom;

  // swap in the new chromosone population
  temp_chrom = first_chrom;
  first_chrom = nfirst_chrom;
  last_chrom = nlast_chrom;

  // Clear up memory of the old chromosone population
  temp_chrom->Kill();

  pop_size = targ_pop_size;
  return(pop_size);   // this->pop_size
}

// Select does shuffle tournament selection, without geneic operators
long Pop::Select(long targ_pop_size)
{
  Shuffle();
  Init_Select(targ_pop_size);
  return(targ_pop_size);   // this->pop_size
}

// Juxta does shuffle tournament selection with cut and splice operators
long Pop::Juxta(long targ_pop_size)
{
  // Do the Cutting and Splicing
  // We're going to try to cut up all of the chromosones
  double cut_chance;
  Chromosone *cur_chrom, *new_chrom, *tail_chrom;
  for(cur_chrom = first_chrom;
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next())
    {
      cut_chance = P_CUT * (((double) cur_chrom->Get_Len() -1.0) /
			    ((double) problem.size));
      if (drand48() < cut_chance)
	{
	  // And now we CUT
	  new_chrom = cur_chrom->Cut(0);

	  // And insert the newly cut chromosone into the population
	  tail_chrom = cur_chrom->Disconnect();
	  cur_chrom->Connect(new_chrom);
	  new_chrom->Connect(tail_chrom);
	  cur_chrom = new_chrom;
	}
    }

  // Then rearrange all of them
  Shuffle();
  
  // Then try to splice all of them back together
  for(cur_chrom = first_chrom;
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next())
    {
      // And we see if we can SPLICE
      if (drand48() < P_SPLICE)
	{
	  if (cur_chrom->Next() != (Chromosone *) NULL)
	    {
	      new_chrom = cur_chrom->Next();
	      tail_chrom = new_chrom->Disconnect();
	      cur_chrom->Splice(new_chrom);
	      cur_chrom->Connect(tail_chrom);
	    }
	}
    }

  // Then do the actual tournament
  Init_Select(targ_pop_size);
  return(pop_size);   // this->pop_size
}

// Return the fitness of the best chromosone
double Pop::Get_Best_Fit()
{
  if (best_chrom == (Chromosone *) NULL)
    return(-1.00);
  else
    return(best_chrom->Get_Fit());
}

// Return the Info of the best chromosone
void Pop::Show_Best(int debug_level)
{
  best_chrom->Show(debug_level);
}
// overloaded for null debug_level
void Pop::Show_Best(void)
{
  best_chrom->Show(0);
}

// Statistics figures average and best values of the population
// returns average fitness value of pop
double Pop::Statistics()
{
  double best_fit, cur_fit, sum_fit;
  Chromosone *cur_chrom;

  best_chrom = first_chrom;
  best_fit = best_chrom->Get_Fit();

  for(cur_chrom = first_chrom->Next();
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next())
    {
      cur_fit = cur_chrom->Get_Fit();
      sum_fit += cur_fit;

      if (cur_fit > best_fit)
	{
	  best_chrom = cur_chrom;
	  best_fit = cur_fit;
	}
    }
  return(sum_fit / pop_size);
}

// Show provides documentation information in mid-flight
void Pop::Show(int debug_level)
{
  if (debug_level < 0)
    return;

  Chromosone *cur_chrom;
  cout << "\nPopulation size: " << pop_size << endl;
  if (debug_level > 10)
    {
      cout << "first_chrom &=" << first_chrom << endl;
      cout << "last_chrom  &=" << last_chrom << endl;
      cout << "tplate  &=" << tplate << endl;
    }
  for(cur_chrom=first_chrom;
      (cur_chrom != (Chromosone *) NULL);
      cur_chrom = cur_chrom->Next())
    {
      cur_chrom->Show(debug_level);
    }
  cout << endl;
}

// Show provides a description of the population (overload d_l=0)
void Pop::Show(void)
{
  Show(0);
}
